#!/usr/bin/php
<?php
/*

soundcloud-dl
Copyright (c) 2013 The soundcloud-dl Author

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

if (count($_SERVER['argv']) < 2) {
    $usage = <<<EOT
Usage: soundcloud-dl [OPTION]...
Download music and metadata from the web service soundcloud.com.

  --get \$mode     Output mode. One of the following strings:
                   title: Write track titles to stdout
                   www:   Write Soundcloud.com URLs to stdout
                   url:   Write best-quality audio URLs to stdout
                   file:  Download files to destination directory
  
  --dest \$path    Output directory for 'file' mode
  --getinfo        Include an information text file in 'file' mode
  
  --max \$int      Maximum number of results to retrieve in successive queries
  
  --user \$name    Query up to \$max of \$user's tracks
  --search \$str   Query up to \$max of the tracks from a search
  --clientid \$str Set client ID for successive requests

  --filter:\$type  Assert that a filter is true for all tracks to be processed.
                   Filters may be one of the following strings:
                   minduration:  Follow with a number of seconds or XXm or XXh
                   namecontains: Follow with a string argument
                   namematch:    Follow with a string argument in PCRE syntax

Homepage: http://code.ivysaur.me/

EOT;
	die($usage);
}

// Defaults:

$max_results = 50;
$pending     = array();
$dest_dir    = './';
$output_mode = 'title';
$getinfo     = false;
define('PUBLIC_CLIENT_ID', 'b45b1aa10f1ac2941910a7f0d10f8e28');

//
// Process arguments
//

$_SERVER['argv'][] = null; // weakly avoid overflowing array bounds
for($i = 1; $i < count($_SERVER['argv']); ++$i) {
	
	switch(strtolower(trim($_SERVER['argv'][$i]))) {
	
		case '--get': {
			$output_mode = ''.trim($_SERVER['argv'][++$i]);
		} break;
		
		case '--getinfo': {
			$getinfo = true;
		} break;
	
		case '--max': {
			$max_results = intval($_SERVER['argv'][++$i]);
		} break;
		
		case '--dest': {
			$dest_dir = rtrim(rtrim(realpath(''.trim($_SERVER['argv'][++$i])), '/'), '\\').'/';
		} break;
	
		case '--clientid': {
			SC::$SUFFIX = 'client_id='.trim($_SERVER['argv'][++$i]);
		} break;
	
		case '--user': {
			$target   = trim(''.$_SERVER['argv'][++$i]);
			$userinfo = SC::resolve($target);
			$tracks   = SC::tracks_for_user($userinfo['id'], $max_results);
			$pending = array_merge($pending, $tracks);
		} break;
		
		case '--search': {
			$target = trim(''.$_SERVER['argv'][++$i]);
			$tracks = SC::search($target, $max_results);
			$pending = array_merge($pending, $tracks);
		} break;
		
		case '--playlist':
		case '--track':
			die("Not implemented\n");
			
		case '--filter:minduration': {
			$min_duration = strtoms( $_SERVER['argv'][++$i] );
			for($ii = 0, $ee = count($pending); $ii != $ee; ++$ii) {
				if ($pending[$ii]['duration'] < $min_duration) unset($pending[$ii]);
			}
			$pending = array_values($pending); // re-key
		} break;
			
		case '--filter:namecontains': {
			$str = strtolower(''.trim($_SERVER['argv'][++$i]));
			for($ii = 0, $ee = count($pending); $ii != $ee; ++$ii) {
				if (strpos(strtolower($pending[$ii]['title']), $str) === false) unset($pending[$ii]);
			}
			$pending = array_values($pending); // re-key
		} break;	
		
		case '--filter:namematch': {
			$str = strtolower(''.trim($_SERVER['argv'][++$i]));
			for($ii = 0, $ee = count($pending); $ii != $ee; ++$ii) {
				if (! preg_match($str, $pending[$ii]['title'])) unset($pending[$ii]);
			}
			$pending = array_values($pending); // re-key
		} break;
			
		case null: break;
		
		default: {
			die("Unrecognised option \"".$_SERVER['argv'][$i]."\"\n");
		} break;
	}
	
}

//
// Parse pending tracks
//

for($i = 0; $i < count($pending); $i++) {
	if (array_key_exists('download_url', $pending[$i]) && strlen($pending[$i]['download_url'])) {
		$pending[$i]['best_url'] = $pending[$i]['download_url'].'?'.SC::$SUFFIX;
	} else if (strlen($pending[$i]['stream_url'])) {
		$pending[$i]['best_url'] = $pending[$i]['stream_url'].'?'.SC::$SUFFIX;
	} else {
		$pending[$i]['best_url'] = null;
	}
}

switch($output_mode) {
	case 'title': {
		$operator = function($track) {
			echo $track['title']."\n";
		};
	} break;
	
	case 'www': {
		$operator = function($track) {
			echo $track['permalink_url']."\n";
		};
	} break;
	
	case 'url': {
		$operator = function($track) {
			if ($track['best_url']) echo $track['best_url']."\n";
		};
	} break;
	
	case 'file': {
		$operator = function($track) use ($dest_dir, $getinfo) {
			$filepath = $dest_dir.filename_esc($track['title']);
			
			// Write information file
			
			if ($getinfo && !file_exists($filepath.'.txt')) {
				file_put_contents(
					$filepath.'.txt',
					
					$track['title']."\r\n".
					"Uploaded ".$track['created_at']."\r\n".
					"\r\n".
					$track['description']."\r\n".
					"\r\n".
					"Tags: ".$track['tag_list']." ".$track['genre']."\r\n".
					"Soundcloud: ".$track['permalink_url']."\r\n"
				);
			}
			
			// Download file
			
			$outform = array_key_exists('download_url', $track) ? $track['original_format'] : 'mp3';			
			if (! file_exists($filepath.'.'.$outform)) {
				
				$in_memory_dl = file_get_contents($track['best_url']);
				file_put_contents($filepath.'.'.$outform, $in_memory_dl);
			}
		};
	} break;
	
	default: {
		die("Unknown output mode \"".$output_mode."\"");
	} break;
}

foreach($pending as $track) {
	$operator($track);
}

//
// Done
//

die();

//
// Helper classes + functions
//

class SC {
	public static $SUFFIX = "";

	public static function resolve($param) {
		return self::get("http://api.soundcloud.com/resolve.json?url=http://soundcloud.com/".urlencode($param));
	}
	
	public static function search($query, $max) {
		return self::get('http://api.soundcloud.com/tracks.json?q='.urlencode($query).'&limit='.intval($max));
	}

	public static function tracks($id) {
		return self::get('http://api.soundcloud.com/tracks/'.intval($id).'.json?');
	}
	
	public static function tracks_for_user($user_id, $max) {
		return self::get('http://api.soundcloud.com/users/'.intval($user_id).'/tracks.json?limit='.intval($max));
	}
	
	private static function get($url) {
		if (! strlen(self::$SUFFIX)) {
			self::$SUFFIX = "client_id=".PUBLIC_CLIENT_ID;
		}
		
		$ret = @file_get_contents($url.'&'.self::$SUFFIX);
		if ($ret === false) {
			die("Not found\n");
		}
		return json_decode($ret, true);
	}
}

function strtoms($str) {
	$str = ''.trim($str);
	$lastchar = strtolower(substr($str, strlen($str)-1, 1));
	$numval = intval($str);
	
	if ($lastchar == 'h') {
		return $numval*60*60*1000;
	} else if ($lastchar == 'm') {
		return $numval*60*1000;
	} else { // seconds
		return $numval*1000;
	}
}

function filename_esc($str) {
	$ret = '';
	for($i = 0, $e = strlen($str); $i < $e; ++$i) {
		if (ctype_alnum($str[$i]) || in_array($str[$i], array('-', ' ', '_', '\''))) {
			$ret .= $str[$i];
		} else {
			$ret .= '_';
		}
	}
	return $ret;
}
